# -------------------------------------------------------------------------------
# Name:        pinkerton_macos.py
# Purpose:     Pinkerton used in MacOS machine.
#
# Author:      Jose Ignacio Palacios Ortega (jipalacios@hispasec.com)
#
# Created:     24/09/2017
# Copyright:   (c) Hispasec Sistemas 2017
# Licence:     <Hispasec Sistemas 2017+>
# -------------------------------------------------------------------------------

# -------------------------------------------------------------------------------
# Import section
# -------------------------------------------------------------------------------

import os
import ssl
import json
import errno
import shutil
import inspect
import zipfile
import argparse
import traceback
import subprocess
import http.client
import configparser
try:
    import tkinter
except (ImportError, ModuleNotFoundError):
    print('\x1b[6;31;01m'
          'Error al importat la librería tk.\n'
          'La librería tk no existe en el sistema.'
          'Instale esta librería y ejecute de nuevo Pinkerton con privilegios '
          'de administrador (root).'
          '\x1b[0m')
    os.sys.exit(1056)

# -------------------------------------------------------------------------------
# Classes and function section
# -------------------------------------------------------------------------------


class Pinkerton(object):
    """ Class that will download Pinkerton source.
    """

    def __init__(self, args):
        """ Class initializer

        Args:
            Namespace with allow following parameters.
                args.verbose: Boolean value to show debug information.
                args.hidden: Boolean value to hidden Pinkerton windows
                args.unninstall: Boolean value to uninstall Pinkerton.
        """

        if args.verbose:
            print('<< %s: %s >>' % (self.__class__.__name__,
                                    inspect.stack()[0][3]))

        self.name = 'pinkerton'
        self.server = 'pinkerton.hispasec.com'
        self.token = '7229b0ee38b6bb4a65979e83f0e3b6ec310b1aa1'
        self.python = args.python

        if not os.path.exists(self.python):
            for python in ('/usr/local/bin/python3',
                           '/usr/bin/python3',
                           '/bin/python3',
                           '/usr/sbin/python3',
                           '/sbin/python3',
                           ('/Library/Frameworks/Python.framework/'
                            'Versions/3.5/bin/python3'),
                           ('/Library/Frameworks/Python.framework/'
                            'Versions/3.6/bin/python3'),
                           ):
                if os.path.exists(python):
                    self.python = python
                    break

        try:
            import requests
        except (ImportError, ModuleNotFoundError):
            try:
                subprocess.call([self.python, '-m', 'pip', 'install',
                                 '--upgrade', 'setuptools'],
                                stdout=subprocess.PIPE)
                subprocess.call([self.python, '-m', 'pip', 'install',
                                 'requests'],
                                stdout=subprocess.PIPE)
            except:
                pass
        self.pinkerton_path = os.path.join(os.environ['HOME'], 'pinkerton')
        self.bin_path = os.path.join(self.pinkerton_path, 'bin')
        self.data_path = os.path.join(self.pinkerton_path, 'data')
        self.font_path = os.path.join('/Library', 'Fonts')
        self.config_file = os.path.join(self.data_path, 'pinkerton.conf')
        self.lock_file = os.path.join(self.pinkerton_path, 'pinkerton.lock')
        self.log_file = os.path.join(self.pinkerton_path, 'pinkerton_log.log')
        self.error_file = os.path.join(self.pinkerton_path,
                                       'pinkerton_error.log')
        self.verbose = args.verbose
        self.hidden = args.hidden
        self.uninstalling = args.uninstall

        if not os.path.exists(self.pinkerton_path):
            os.mkdir(self.pinkerton_path)
            if self.verbose:
                log_text = 'Directorio creado: %s\n' % self.pinkerton_path
                print(log_text)
                with open(self.log_file, 'at') as log_flie:
                    log_flie.write('<< %s: %s >>' % (self.__class__.__name__,
                                                     inspect.stack()[0][3]))
                    log_flie.write('%s\n' % log_text)

        if not os.path.exists(self.bin_path):
            os.mkdir(self.bin_path)
            if self.verbose:
                log_text = 'Directorio creado: %s' % self.bin_path
                print(log_text)
                with open(self.log_file, 'at') as log_flie:
                    log_flie.write('%s\n' % log_text)

        if not os.path.exists(self.data_path):
            os.mkdir(self.data_path)
            if self.verbose:
                log_text = 'Directorio creado: %s' % self.data_path
                print(log_text)
                with open(self.log_file, 'at') as log_flie:
                    log_flie.write('%s\n' % log_text)

        if not os.path.exists(self.config_file):
            config = configparser.ConfigParser()
            config['SERVER'] = {'server': self.server,
                                'token': self.token}
            with open(self.config_file, 'wt') as cf:
                config.write(cf)

            if self.verbose:
                log_text = 'Fichero creado: %s' % self.config_file
                print(log_text)
                with open(self.log_file, 'at') as log_flie:
                    log_flie.write('%s\n' % log_text)

        if os.path.exists(self.error_file):
            os.remove(self.error_file)

        if os.path.exists(self.log_file):
            os.remove(self.log_file)

        if not os.path.exists(self.error_file):
            with open(self.error_file, 'wt') as f:
                f.write(str())

        if not os.path.exists(self.log_file):
            with open(self.log_file, 'wt') as f:
                f.write(str())

    def run(self):
        """ Launch install process
        This will update Pinkerton software and launch it.

        Post:
            Pinkerton will be installed correctly.
            Pinkerton will be updated.
            Pinkerton will be launched.
        """

        if self.verbose:
            log_text = ('<< %s: %s >>' % (self.__class__.__name__,
                                          inspect.stack()[0][3]))
            print(log_text)
            with open(self.log_file, 'at') as log_flie:
                log_flie.write('%s\n' % log_text)
                log_flie.write('Parámetros actuales:\n%s\n ' %
                               json.dumps(self.dump(), indent=2))

        try:
            if self.verbose:
                log_text = 'Comprobando privilegos del usuario...'
                print(log_text)
                with open(self.log_file, 'at') as log_flie:
                    log_flie.write('%s\n' % log_text)

                with open(os.path.join(self.font_path, 'pinkerton.ttf'), 'w') \
                        as f:
                    f.write(str())

                os.remove(os.path.join(self.font_path, 'pinkerton.ttf'))

            libs = os.path.join(self.bin_path, 'pinkerton.zlibs')
            if self.verbose:
                log_text = ('Usuario con permisos adecuados.\n'
                            'Conectando con el servidor...')
                print(log_text)
                with open(self.log_file, 'at') as log_flie:
                    log_flie.write('%s\n' % log_text)

            try:
                conn = http.client.HTTPSConnection(self.server)
                conn.request('GET', '/pinkerton/update/')
            except:
                try:
                    conn = http.client.HTTPSConnection(
                        self.server, context=ssl._create_unverified_context())
                    conn.request('GET', '/pinkerton/update/')
                except:
                    raise ConnectionError

            if self.verbose:
                log_text = ('Conextión establecida.\n'
                            'Hacienddo la petición al servidor...')
                print(log_text)
                with open(self.log_file, 'at') as log_flie:
                    log_flie.write('%s\n' % log_text)

            response = conn.getresponse()
            if response.status == 200:
                if self.verbose:
                    log_text = ('Petición realizaza satisfactoriamente.\n'
                                'Guardando la información recibida...')
                    print(log_text)
                    with open(self.log_file, 'at') as log_flie:
                        log_flie.write('%s\n' % log_text)

                with open(libs, 'wb')as f:
                    f.write(response.read())

                if self.verbose:
                    log_text = ('Información guardada satisfactoriamente.\n'
                                'Extrayendo la información guardada...')
                    print(log_text)
                    with open(self.log_file, 'at') as log_flie:
                        log_flie.write('%s\n' % log_text)

                with zipfile.ZipFile(libs) as zf:
                    if not(os.path.exists(os.path.join(self.font_path,
                                                       'D3Euronism.ttf'))):
                        with open(os.path.join(self.font_path,
                                               'D3Euronism.ttf'), 'wb') as f:
                            f.write(zf.read('D3Euronism.ttf'))

                    if not(os.path.exists(os.path.join(self.font_path,
                                                       'Roboto-Regular.ttf'))):
                        with open(os.path.join(self.font_path,
                                               'Roboto-Regular.ttf'),
                                  'wb') as f:
                            f.write(zf.read('Roboto-Regular.ttf'))

                    for f in zf.filelist:
                        if f.filename.split('.')[-1] == 'py':
                            with open(os.path.join(self.bin_path, f.filename),
                                      'wb') as fout:
                                fout.write(zf.read(f.filename))

                if self.verbose:
                    log_text = ('Información extraída satisfactoriamente.\n'
                                'Registrando la aplicación...')
                    print(log_text)
                    with open(self.log_file, 'at') as log_flie:
                        log_flie.write('%s\n' % log_text)

                kernel = os.path.join(self.bin_path, 'kernel.py')
                command = [self.python, kernel]
                if self.verbose:
                    command.append('-v')

                if self.hidden:
                    command.append('--hidden')

                if self.verbose:
                    log_text = ('Aplicación registrada satisfactoriamente.\n'
                                'Ejecutando Pinkerton...')
                    print(log_text)
                    with open(self.log_file, 'at') as log_flie:
                        log_flie.write('%s\n' % log_text)

                subprocess.call(command)
            else:
                log_text = 'Ocurrió un error al conectar con el servidor'
                print('\x1b[6;31;01m%s \x1b[0m' % log_text)
                with open(self.log_file, 'at') as log_flie:
                    log_flie.write('%s\n' % log_text)

                with open(self.error_file, 'at') as log_flie:
                    log_flie.write('%s\n' % log_text)

            self.clean()

        except KeyboardInterrupt:
            log_text = 'Aplicación parada por el usuario'
            print('\x1b[6;33;01m%s \x1b[0m' % log_text)
            with open(self.log_file, 'at') as log_flie:
                log_flie.write('%s\n' % log_text)

            self.clean()

        except Exception:
            log_text = 'Ocurrió una excepción'
            print('\x1b[6;31;01m%s \x1b[0m' % log_text)
            log_text = '\n'.join(traceback.format_exc().splitlines())

            with open(self.log_file, 'at') as log_flie:
                log_flie.write('%s\n' % log_text)

            with open(self.error_file, 'at') as log_flie:
                log_flie.write('%s\n' % log_text)

            self.clean()
            raise

    def clean(self):
        """ Funtion to clean Pinkerton directory.

        Post.
            All files that all unnecesary will be deleted
        """

        if self.verbose:
            log_text = ('<< %s: %s >>' % (self.__class__.__name__,
                                          inspect.stack()[0][3]))
            print(log_text)
            with open(self.log_file, 'at') as log_flie:
                log_flie.write('%s\n' % log_text)

        libs = os.path.join(self.bin_path, 'pinkerton.zlibs')
        try:
            with zipfile.ZipFile(libs) as zf:
                for fileinfo in zf.filelist:
                    try:
                        os.remove(os.path.join(self.bin_path,
                                               fileinfo.filename))
                    except IOError as exception:
                        if not exception.errno == errno.ENOENT:
                            if self.verbose and self.uninstalling:
                                log_text = ('El fichero %s no se encuentra en '
                                            'el sistema' % fileinfo.filename)
                                print('\x1b[6;33;01m%s \x1b[0m' % log_text)
                                with open(self.log_file, 'at') as log_flie:
                                    log_flie.write('%s\n' % log_text)
                    except Exception:
                        pass

        except IOError as exception:
            if exception.errno == errno.ENOENT:
                if self.verbose and self.uninstalling:
                    log_text = ('El fichero %s  no se encuentra en el sistema'
                                % libs)
                    print('\x1b[6;31;01m%s \x1b[0m' % log_text)
                    with open(self.log_file, 'at') as log_flie:
                        log_flie.write('%s\n' % log_text)

                    with open(self.error_file, 'at') as log_flie:
                        log_flie.write('%s\n' % log_text)
            else:
                raise

        except Exception as why:
            log_text = str(why)
            with open(self.log_file, 'at') as log_flie:
                log_flie.write('%s\n' % log_text)

            with open(self.error_file, 'at') as log_flie:
                log_flie.write('%s\n' % log_text)

            raise

    def uninstall(self):
        """ Uninstall Pinkerton.

        Post:
            Pinkerton will be uninstalled and all file will be deleted.
        """

        if self.verbose:
            log_text = ('<< %s: %s >>' % (self.__class__.__name__,
                                          inspect.stack()[0][3]))
            print(log_text)
            with open(self.log_file, 'at') as log_flie:
                log_flie.write('%s\n' % log_text)

        try:
            if os.path.exists(os.path.join(self.font_path, 'D3Euronism.ttf')):
                os.remove(os.path.join(self.font_path, 'D3Euronism.ttf'))

            if os.path.exists(os.path.join(self.font_path,
                                           'Roboto-Regular.ttf')):
                os.remove(os.path.join(self.font_path, 'Roboto-Regular.ttf'))

            if os.path.exists(self.pinkerton_path):
                shutil.rmtree(self.pinkerton_path)

        except IOError:
            self.clean()
            raise

    def dump(self):
        """ Function to get class attributes

        Return:
            data: Dictionary with class attributes.
        """

        if self.verbose:
            print('<< %s: %s >>' % (self.__class__.__name__,
                                    inspect.stack()[0][3]))
            with open(self.log_file, 'at') as log_flie:
                log_flie.write('<< %s: %s >>' % (self.__class__.__name__,
                                                 inspect.stack()[0][3]))

        data = {'name': self.name,
                'server': self.server,
                'python': self.python,
                'pinkerton_path': self.pinkerton_path,
                'bin_path': self.bin_path,
                'data_path': self.data_path,
                'font_path': self.font_path,
                'lock_file': self.lock_file,
                'log_file': self.log_file,
                'error_file': self.error_file,
                'verbose': self.verbose,
                'hidden': self.hidden,
                'uninstalling': self.uninstalling,
                }

        return data

# -------------------------------------------------------------------------------
# Main function
# -------------------------------------------------------------------------------


if __name__ == '__main__':
    # La maquina ha de tener tkinter para compilar correctamente pinkerton
    # prerequisitos requests "sudo python3 -m pip install requests"
    parser = argparse.ArgumentParser(prog='Pinkerton')
    parser.add_argument('-v', '--verbose', default=False, action='store_true',
                        help='Muestra información de depuración')
    parser.add_argument('--hidden', default=False, action='store_true',
                        help='No muestra la interfaz de Pinkerton')
    parser.add_argument('-p', '--python', default=str(),
                        help='Binario de python')
    parser.add_argument('-u', '--uninstall', default=False, action='store_true',
                        help='Argument to uninstall Pinkerton')
    main_args = parser.parse_args()

    if main_args.verbose:
        print('%s\n<< Lanzando >>\n%s' % ('*' * 32, '*' * 32))

    try:
        pinkerton = Pinkerton(main_args)
        if not main_args.uninstall:
            pinkerton.run()

            if main_args.verbose:
                print('\x1b[6;34;01m'
                      'Aplicación terminada correctamente.'
                      '\x1b[0m')
        else:
            pinkerton.uninstall()

            if main_args.verbose:
                print('\x1b[6;34;01m'
                      'Aplicación desinstalada correctamente.'
                      '\x1b[0m')

    except IOError as reason:
        if reason.errno == errno.EACCES:
            print('\x1b[6;31;01m'
                  'Permisos insuficientes para lanzar Pinkerton.'
                  '\nEjecute Pinkerton con permisos de administrador.'
                  '\x1b[0m')
        else:
            print(str(reason))

    except ConnectionError:
        print('\x1b[6;31;01m'
              'Error al intentar conectar con el servidor de Pinkerton.\n'
              'Asegúrese que la red funciona correctamente.'
              '\x1b[0m')

    except Exception as reason:
        print('\x1b[6;31;01m' + str(reason) + '\x1b[0m')

    if main_args.verbose:
        print('%s\n<< Terminado >>\n%s' % ('*' * 32, '*' * 32))
